open Format

module type COEF = sig
  type t
  val of_int : string -> t
  val zero : t
  val one : t 
  val minusone : t 
  val plus : t -> t -> t
  val minus : t -> t -> t
  val mult : t -> t -> t
  val inverse : t -> t

  val compare : t -> t -> int

  val print : Format.formatter -> t -> unit

end

module type S = sig

  type coef

  exception Unsolvable
  type dumview = coef Term.Map.t
  type view = coef Term.Map.t * coef
  type t

  module Set : Set.S with type elt = t
  module Map : Map.S with type key = t

  val is_num : t -> bool
  val is_var : t -> Term.t Sig.mine

  val make : Term.t -> t

  val leaves : t -> Term.Set.t

  val view : t -> view
  val dumview : t -> dumview

  val compare : t -> t -> int

  val mult : coef -> t -> t
  val add : t -> t -> t

  val choose : t -> Term.t * coef
  val subst : Term.t -> t -> t -> t
  val canon : t -> (Term.t -> t) -> t 
  val solve : t -> t -> (Term.t * t) list

  val print : Format.formatter -> t -> unit

end

module Make (C:COEF) = struct

  type coef = C.t

  exception Unsolvable
  exception Notlinear

  module T = Term
  module S = Symbols
  module Mterm = T.Map

  type dumview = coef Term.Map.t
  type view = coef Mterm.t * coef
  type t = coef Mterm.t * coef

  let print fmt (m,v) =
    Mterm.iter 
      (fun t n -> fprintf fmt "%a*%a " C.print n T.print t) m;
    fprintf fmt "%a" C.print v

  let is_num (m,_) = Mterm.is_empty m
  let is_n v n = C.compare n v = 0
  let is_var (m,v) = 
    if is_n v C.zero then
      (let r = ref [] in
       try Mterm.iter 
	 (fun t v -> 
	    if List.length !r > 1 then raise Exit;
	    if not (is_n v C.zero) then
	      if is_n v C.one then r:=t::!r else raise Exit) m;
	 (match !r with
	      [t] -> Sig.No t
	    | _ -> Sig.Yes)
       with Exit -> Sig.Yes)
    else Sig.Yes

  let ep = Mterm.empty,C.zero

  let find t m = try Mterm.find t m with Not_found -> C.zero

  let mult_map m v k =  Mterm.map (fun v->C.mult k v) m , C.mult k v

  let merge_map = 
    Mterm.fold 
      (fun x v m -> 
	 let coef = C.plus (find x m) v in
	 if C.compare coef C.zero = 0 then Mterm.remove x m
	 else Mterm.add x (C.plus (find x m) v) m)

  let rec mke coef ((m,v) as acc) t f = try
    let m1,v1 = Term.Map.find t f in
    let m2,v2 = mult_map m1 v1 coef in
    merge_map m2 m , C.plus v2 v
  with Not_found ->
    let {T.f=sb ; xs=xs} = T.view t in
    match sb , xs with
      | S.Int n , _  -> m , C.plus (C.mult coef (C.of_int (Hstring.view n))) v
      | S.Binop S.Mult, [t1;t2] ->
	  let (m1,v1) = mke C.one ep t1 f in
	  let (m2,v2) = mke C.one ep t2 f in
	  if Mterm.is_empty m1 then
	    if Mterm.is_empty m2 then 
	      m, C.plus v (C.mult (C.mult coef v1) v2)
	    else 
	      if C.compare v1 C.zero = 0 then m,v 
	      else 
		let m2,v2 = mult_map m2 v2 (C.mult coef v1) in
		merge_map m2 m , C.plus v2 v
	  else 
	    if Mterm.is_empty m2 then
	      if C.compare v2 C.zero = 0 then m,v 
	      else 
		let m1,v1 = mult_map m1 v1 (C.mult coef v2) in
		merge_map m1 m , C.plus v1 v
	    else 
	    let c = try Mterm.find t m with Not_found -> C.zero in
	    Mterm.add t (C.plus coef c) m , v
      | S.Binop S.Div, [t1;t2] -> 
	  let c = try Mterm.find t m with Not_found -> C.zero in
	  Mterm.add t (C.plus coef c) m , v
	  (*raise Notlinear*)
	  (*	    let (m1,v1) = mke cun ep t1 f in
		    let (m2,v2) = mke cun ep t2 f in
		    if Mterm.is_empty m1 && Mterm.is_empty m2 && 
		    C.compare v2 czero <>0 then 
		    m, C.plus v (C.mult coef (v1/v2))
		    else raise Notlinear*)
      | S.Binop S.Plus , [t1;t2] -> 
	  mke coef (mke coef acc t2 f) t1 f
      | S.Binop S.Minus , [t1;t2] -> 
	  mke coef (mke (C.minus C.zero coef) acc t2 f) t1 f
      | S.Name _ , _ -> 
	  let c = C.plus (find t m) coef in
	  if C.compare c C.zero = 0 then Mterm.remove t m , v 
	  else Mterm.add t c m , v
      | _ -> eprintf "%a @." S.print sb; assert false
	  
  let make t = mke C.one ep t Term.Map.empty
(*    match T.view t with
	{T.ty=Ty.Tint} -> mke C.one ep t Term.Map.empty
      | _ -> raise Not_found*)
      
  let view x = x
  let dumview  (m,v) = Mterm.add Term.dummy v m

  let leaves (m,v) = Mterm.fold (fun t _ -> Term.Set.add t) m Term.Set.empty
    
  let compare (m1,a1) (m2,a2) = 
    let c = Mterm.compare C.compare m1 m2 in
    if c=0 then C.compare a1 a2 else c

  module STR = Set.Make(
    struct
      type t' = Term.t * t
      type t = t'
      let compare (t1,r1) (t2,r2) = 
	let c = Term.compare t1 t2 in
	if c<>0 then compare r1 r2 else c
    end)

  module Set = 
    Set.Make(struct type t'=t type t=t' let compare=compare end)

  module Map = 
    Map.Make(struct type t'=t type t=t' let compare=compare end)

  let mult n (m,a) = Mterm.map (C.mult n) m , C.mult n a
      
  let add (m1,a1) (m2,a2) = 
    let m = 
      Mterm.fold 
	(fun x a m -> 
	   let v = try Mterm.find x m with Not_found -> C.zero in
	   let a' = C.plus v a in
	   if C.compare a' C.zero = 0 then 
	     Mterm.remove x m else Mterm.add x a' m)
	m2 m1 
    in m , C.plus a1 a2
      
  let choose (m,_) = 
    let tn= ref None in
    (try Mterm.iter (fun k v -> tn:=Some(k,v); raise Exit) m
     with Exit -> ());
    match !tn with
	Some p -> p
      | _ -> raise Not_found

  let subst x r1 ( (m2,v2) as r2) =
    try
      let a = Mterm.find x m2 in
      add (mult a r1) (Mterm.remove x m2,v2)
    with Not_found -> r2

  let canon (m,v) f = 
    Mterm.fold 
      (fun x a poly -> 
	 let r = 
	   try f x 
	   with Not_found -> 
	     (Mterm.add x C.one Mterm.empty), C.zero
	 in
	 add (mult a r) poly
      )
      m (Mterm.empty,v)

  let solve r1 r2 = 
    let ((m,v) as r) = add r1 (mult (C.of_int "-1") r2) in
    try 
      let x , a = choose r in
      [x , mult (C.inverse a) (Mterm.remove x m,v)]
    with Not_found -> if C.compare v C.zero <> 0 then raise Unsolvable; []
	
end

