open Options
open Format

module A = Atom
module D = Cc.Make(Cc.CombineX)
module F = Formula
module M = Matching.Make(D) 
module SF = F.Set
module MF = F.Map

type gformula = F.t * int * F.t option

type t = { 
    gamma : SF.t ; 
    delta : (gformula * gformula) list;
    tbox : D.t;
    lemmas : int MF.t;
    matching : M.t
}
      
exception Sat
exception Unsat
exception I_dont_know

(* matching part of the solver *)

let add_terms env s age lem = 
  { env with 
      matching = Term.Set.fold (M.add age lem) s env.matching }

let mtriggers env lem age acc = 
  match F.view lem with
      F.L {F.trs=tgs;F.f=f;name=name} ->
	List.fold_left 
	  (fun acc pat ->
	     List.fold_left 
	       (fun acc {Matching.sbt=s;gen=g} -> 
		  let nf = F.apply_subst s f in
		  if SF.mem nf env.gamma then acc else 
		    (if debug_sat_simple then printf "%s@." name; 
		     (nf,1+(max g age),Some lem)::acc)
	       )
	       acc (M.matching lem pat env.matching env.tbox))
	  acc tgs
    | _ -> assert false

let mround env = MF.fold (mtriggers env) env.lemmas []

(* sat-solver *)

let elim (f,_,_) env = 
  SF.mem f env.gamma ||
    match F.view f with F.A a -> D.query a (D.add a env.tbox)  | _ -> false

let red (f,a,b) env = elim ((F.mk_not f),a,b) env

let rec assume env (f,age,lem) = 
  try
    if SF.mem (F.mk_not f) env.gamma then raise Unsat;
    if SF.mem f env.gamma then env
    else 
      begin
	let env = if glouton then add_terms env (F.terms f) age lem else env in
	let env = { env with gamma = SF.add f env.gamma } in
	match F.view f with
	    F.U l -> 
	      List.fold_left assume env (List.map (fun x->x,age,lem) l)
	  | F.C(f1,f2) -> 
	      if debug_sat then	
		printf "[sat] I assume a clause %a@." F.print f;
	      bcp { env with delta = ((f1,age,lem),(f2,age,lem))::env.delta }
	  | F.L _ -> 
	      if debug_sat then printf "[sat] I assume a lemma %a@." F.print f;
	      let age = 
		try min age (MF.find f env.lemmas) with Not_found -> age in
	      bcp { env with lemmas=MF.add f age env.lemmas }
	  | F.A a -> 
	      let a = D.normal_form a in
	      if debug_sat then	
		printf "[sat] I assume an atom %a@." Atom.print a;
	      let env = add_terms env (A.terms_of a) age lem in
	      let proof = SF.singleton f in
	      let env = { env with tbox = D.assume a proof env.tbox } in
	      bcp env
	  | F.Sko(s,f) -> 
	      let f' = F.apply_subst (s,Ty.esubst) f in
	      if debug_sat then 
		printf "[sat] I assume a skolem %a [%a]@." 
		  F.print f F.print f';
	      assume env (f',age,lem)
      end
      with D.Inconsistent -> raise Unsat 
	
and bcp env = 
  let cl , u = 
    List.fold_left 
      (fun (cl,u) (f1,f2) -> 
	 if elim f1 env || elim f2 env  then (cl,u)
	 else if red f1 env then (cl,f2::u)
	 else if red f2 env then (cl,f1::u)
	 else (f1,f2)::cl , u
      ) ([],[]) env.delta
  in
  List.fold_left assume {env with delta=cl} u
    
let rec unsat env fg stop = 
  try
    if stop < 0 then raise I_dont_know;
    back_tracking (assume env fg) stop
  with Unsat -> ()

and back_tracking env stop = match env.delta with
    []  when stop >= 0  -> 
      (match mround env with
	   [] -> raise Sat
	 | l -> back_tracking (List.fold_left assume env l) (stop-1))
  | [] -> raise I_dont_know
  | ((f,g,lem) as a,b)::l -> 
      unsat {env with delta=l} a stop;
      unsat (assume {env with delta=l} b) (F.mk_not f,g,lem)  stop

let empty = { 
  gamma = SF.empty;
  delta = [] ;
  tbox = D.empty ; 
  lemmas = MF.empty ; 
  matching = M.empty
} 




