open Format
open Options

module type S = sig
  type t
    
  module R : Sig.THEORY
  module Res : Set.S with type elt = R.t * Term.Set.t * R.Set.t

  exception Unknown_element
  exception Inconsistent

  val empty :  t
  val add : Term.t -> t -> t
  val mem : Term.t -> t -> bool
  val find : t -> Term.t -> R.t
  val union : t -> Term.t -> Term.t -> Formula.Set.t -> bool -> t * Res.t
  val distinct : t -> Term.t -> Term.t -> Formula.Set.t -> t

  val leaves : t -> Term.t -> R.Set.t
  val equal : t -> Term.t -> Term.t -> bool
  val are_distinct : t -> Term.t -> Term.t -> bool
  val classof : t -> Term.t -> Term.t list

  val explain : t -> Term.t -> Term.t -> Formula.Set.t
  val nexplain : t -> Term.t -> Term.t -> Formula.Set.t

end

module Make ( X : Sig.THEORY ) = struct

  module R = X
  module S = Symbols
  module T = Term
  module MapT = Term.Map
  module SetT = Term.Set
  module SF = Formula.Set
    
  module TSF = Set.Make(
    struct
      type t = Term.t * SF.t
      let compare (t1,_) (t2,_) = Term.compare t1 t2
    end)

  module Res = Set.Make(
    struct
      type t = R.t * Term.Set.t * R.Set.t
      let compare (r1,st1,sr1) (r2,st2,sr2) = 
	let c = R.compare r1 r2 in 
	if c<>0 then
	  let c = Term.Set.compare st1 st2 in
	  if c<>0 then
	    R.Set.compare sr1 sr2
	  else c
	else c
    end)

  type t = { 
    map : ((R.t * SF.t) list) MapT.t ; 
    (* associates each term to its representative *)
    
    minv : SetT.t R.Map.t ;  
    (* associates each representative to its class (of terms) *)
    
    mapm : SetT.t MapT.t; 
    (* associates each monome m to the set of terms whose 
       representative contains m *)
    
    neqs: TSF.t R.Map.t;
  }
      
  let afind env x = List.hd (MapT.find x env.map)
  let rfind env x = fst (afind env x)
  let efind env x = snd (afind env x)

  exception Unknown_element
  exception Inconsistent
    
  let empty = { 
    map  = MapT.empty; 
    minv = R.Map.empty; 
    mapm = MapT.empty;
    neqs = R.Map.empty;
  }
  
  let print_map fmt m = 
    let print_set fmt s =
      SF.iter 
        (fun x -> fprintf fmt "%a, " Formula.print x) s
    in
(*    MapT.iter (fun k l -> try
		   let x , _ = List.hd l in
		   fprintf fmt "%a -> %a@. " T.print k R.print x
	       with Not_found -> fprintf fmt "%a ->@." T.print k ) m*)
      fprintf fmt "@.";
      MapT.iter 
        (fun k l ->
           fprintf fmt "[ ";
           List.iter (fun (v,d) -> 
             fprintf fmt "%a -> %a { %a } ;@," T.print k R.print v print_set d)
             l;
           fprintf fmt "]@. "
             ) m
        
  let sa_print fmt = SF.iter (fprintf fmt "%a" Formula.print)

  let tsa_print fmt = 
    TSF.iter 
      (fun (t,sa) -> fprintf fmt "(%a:%a)" T.print t sa_print sa) 

  let print_neqs fmt m = 
    fprintf fmt "@.";
    R.Map.iter 
      (fun k s -> fprintf fmt "%a|->%a\n" R.print k tsa_print s) m;
    fprintf fmt "@."

  let add_mapm t p mapm = 
    let s = R.leaves p in
    SetT.fold 
      (fun x mapm -> 
	 let s = try MapT.find x mapm with Not_found -> SetT.empty in
	 MapT.add x (SetT.add t s) mapm) s mapm
    
  let add_minv v mv minv = 
    let s = 
      try SetT.add v (R.Map.find mv minv)
      with Not_found -> (SetT.singleton v)
    in
    R.Map.add mv s minv
    
  let add_mneqs s r proof env = 
    if TSF.exists 
      (fun (u,_) -> R.compare r (rfind env u) = 0) s
    then raise Inconsistent;
    let s' = 
      let s = 
	TSF.fold 
	  (fun (t,ex) s -> TSF.add (t,SF.union ex proof) s) s TSF.empty
      in
      try TSF.union s (R.Map.find r env.neqs) with Not_found -> s
    in
    R.Map.add r s' env.neqs
    
  let init_neqs r neqs = 
    if R.Map.mem r neqs then neqs
    else R.Map.add r TSF.empty neqs
    
  let canon r env =
    SetT.fold 
      (fun x newr -> 
	 let rx = try rfind env x with Not_found -> R.make x in
	 R.subst x rx newr) (R.leaves r) r
      
  let add t env = 
    if MapT.mem t env.map then env
    else 
      let r = canon (R.make t) env in
      { map  = MapT.add t [(r,SF.empty)] env.map ; 
	minv = add_minv t r env.minv;
	mapm = add_mapm t r env.mapm;
	neqs = init_neqs r env.neqs;
      }

  let mem t env = MapT.mem t env.map
      
  let find env t = 
    let res =
      try rfind env t with Not_found -> (* assert false*) R.make t  in
      if debug_uf then printf "find: %a -> %a@." T.print t R.print res;
      res
    
  let classof env v = try 
    SetT.elements (R.Map.find (rfind env v) env.minv)
  with Not_found -> [v]

  let one = R.make (Term.make (Symbols.name "@bottom") [] Ty.Tint) 

  let leaves_r env r = try
    let s = R.leaves r in
    if SetT.is_empty s then R.Set.singleton one
    else
      SetT.fold (fun x l -> R.Set.add (rfind env x) l) s R.Set.empty
  with Not_found -> assert false 
    
  let leaves env t = 
    try leaves_r env (rfind env t) with Not_found -> assert false
      
  let update x r env proof = 
    try
      SetT.fold 
	(fun t env -> 
           let h = MapT.find t env.map in
	   let rt , dt = List.hd h in
	   let nrt = R.subst x r rt in
	   let neqs_rt = R.Map.find rt env.neqs in
	   { map = MapT.add t ((nrt,SF.union proof dt)::h) env.map;
	     mapm = add_mapm t nrt env.mapm;
	     minv = add_minv t nrt env.minv ;
	     neqs = add_mneqs neqs_rt nrt proof env;
	   }
	) 
	(* il faut faire le menage dans les maps *)
	(MapT.find x env.mapm) env
    with Not_found -> assert false

  let union env x y proof ccb =
    if debug_uf then printf "[uf] union %a = %a@." T.print x T.print y;
    let env = add y (add x env) in
    let rx , dx = afind env x in
    let ry , dy = afind env y in 
    if R.compare rx ry = 0 then env , Res.empty
    else
      let () = Traces.add_eq x y ccb in
      try
	if TSF.mem (y,SF.empty) (R.Map.find rx env.neqs) then
	  raise Inconsistent;
	if TSF.mem (x,SF.empty) (R.Map.find ry env.neqs) then
	  raise Inconsistent;
	let env,res = 
	  List.fold_left
	    (fun (env,res) (t,v) ->
	       let ro =  rfind env t in
	       if debug_uf then
		 printf "[uf] on pivote sur %a (soit %a)@." 
		   T.print t R.print ro;
	       let touched = MapT.find t env.mapm in
	       let d = SF.union dx dy in
	       let env = update t v env (SF.union proof d) in
	       env , Res.add (ro,touched,leaves_r env v) res) 
	    (env,Res.empty) (R.solve rx ry)
	in
	if debug_uf then 
	  (printf "map de uf: %a@." print_map env.map;
	   printf "neq map de uf: %a@." print_neqs env.neqs);
	env , res
      with R.Unsolvable -> raise Inconsistent
	
  let distinct m x y proof = 
    if debug_uf then printf "[uf] distinct %a <> %a@." 
      T.print x T.print y;
    let m = add y (add x m) in
    let rx = rfind m x in
    let ry = rfind m y in 
    if R.compare rx ry = 0 then 
      (if tracefile <> "" then Traces.print_trace x y; raise Inconsistent);
    let dx = R.Map.find rx m.neqs in
    let dy = R.Map.find ry m.neqs in
    { m with 
	neqs = 
	R.Map.add rx (TSF.add (y,proof) dx) 
	  (R.Map.add ry (TSF.add (x,proof) dy) m.neqs) }
      
  let equal env t1 t2 = try
    R.compare (rfind env t1) (rfind env t2) = 0
  with Not_found ->  
    let r1 = canon (R.make t1) env in
    let r2 = canon (R.make t2) env in
    R.compare r1 r2 = 0

  let are_distinct m t1 t2 = 
    let b= 
      try
	match T.view t1 , T.view t2 with
	    {T.f=S.Int n1} , {T.f=S.Int n2} -> Hstring.compare n1 n2 <> 0
	  | _ -> 
	      let dt1 = R.Map.find (find m t1) m.neqs in
	      let dt2 = R.Map.find (find m t2) m.neqs in
	      TSF.mem (t1,SF.empty) dt2 || TSF.mem (t2,SF.empty) dt1
      with Not_found -> false
    in     
    if debug_uf then
      printf "[ufarith] are_distinct %a<>%a ? %b@." T.print t1 T.print t2 b; 
    b

  let common_repr h1 h2 =
    let rec aux h1 h2 e1 e2 =
      match h1,h2 with
        | [],_ | _,[] -> e1, e2
        | ((r1,ex1)::xs), ((r2,ex2)::ys) ->
            if R.compare r1 r2 = 0 then
              aux xs ys ex1 ex2
            else
              e1,e2
    in
      aux h1 h2 (snd (List.hd h1)) (snd (List.hd h2))

  let explain env x y = 
    let h1 = MapT.find x env.map in
    let h2 = MapT.find y env.map in
    let e1, e2 = common_repr h1 h2 in
      SF.union e1 e2

  let nexplain env x y = 
    let rx , dx = afind env x in
    let ry , dy = afind env y in
    let s_rx = R.Map.find rx env.neqs in
    let s_ry = R.Map.find ry env.neqs in
    let s1 = try 
      snd(TSF.choose (TSF.filter (fun (t,_) -> Term.equal t x) s_ry)) 
    with Not_found -> SF.empty
    in
    let s2 = try
      snd(TSF.choose (TSF.filter (fun (t,_) -> Term.equal t y) s_rx)) 
    with Not_found -> SF.empty
    in
    SF.union s1 (SF.union s2 (SF.union dx dy))
      
end
